;==========================================================================
; File Name   :	z8671.rom
;==========================================================================
; reassembled and commented Volker Pohlers, 10/2021, Germany, Neustadt i.H.
;
; das Z8671-BASIC/DEBUG ist eine TINY-BASIC-Implementierung mit eigener
; Interpreter Sprache IL, wie in [1] vorgestellt.
; Die IL-Codes stimmen grtenteils mit denen berein, die Tom Pittman
; in [2] beschrieben hat.
;
; [1] Dr. Dobb's Journal, Volume 1, Number 1, 1976, p. 12.
; [2] http://www.ittybittycomputers.com/IttyBitty/TinyBasic/TBEK.txt
;==========================================================================
;f. Arnold-Assembler
;AS-Funktionen
hi              function x,(x>>8)&255
lo              function x, x&255

	cpu	z8601
	assume RP:0C0h		; keine Optimierung !
;==========================================================================

; anpassung an ES40
; Anpassung an ES1988

		org	8000h

; irq0:		dw 1000h + start
; irq1:		dw 1003h + start
; irq2:		dw 1006h + start
; irq3:		dw 1009h + start
; irq4:		dw 100Ch + start
; irq5:		dw 100Fh + start

		; public start
start:		jp	RESET
		jp	load
		jp	save
		jp	load0		; Programmende setzen
	
; Registers: RP=10h
; -----------------
; RR2  return value, expr
; RR6  Basic line number
; RR8  IL PC
; RR10 Basic Pgm. Pointer
; R15  IL Stack Pointer

;------------------------------------------------------------------------------
; pop the TOS word of stack -> RR2
;------------------------------------------------------------------------------

IL_SP:		ld	R3, @R15
		inc	R15
		ld	R2, @R15
		inc	R15
		ret

;------------------------------------------------------------------------------
;  IL instruction print	tab
;------------------------------------------------------------------------------

IL_PT:		ld	R3, #' '	; emit a space
		call	output
		tm	cupos, #7	; Print	column counter,	cursor position
		jr	NZ, IL_PT
		ret

;------------------------------------------------------------------------------
; 0020 default constant block
;------------------------------------------------------------------------------
; must be a xxx0-address!
; max block size 15 Bytes

		align 10h

dcblk:		db  hi(il)	; 0  00650h
		db  lo(il) 	; 1
		db  20h		; 2  stack reserve
		db  8		; 3  backspace code
		db  1Bh		; 4  cancel code
		db  0Dh		; 5  line separator sequence terminating with 0FEh
;;		db  0Ah		; 6

;------------------------------------------------------------------------------
; End of IL_GL
;------------------------------------------------------------------------------

loc_27:		inc	R15		; restore stack (see IL_GL)
		call	loc_81		; @RR4 = 0, set end of basic line

;------------------------------------------------------------------------------
; IL instruction: emit new line
;------------------------------------------------------------------------------

IL_NL:		ld	R3, #5		; line separator sequence
		call	get_cblk	; get from constant block
loc_30:		call	output1
		call	get_cblk1	; next byte from sequence
		jr	NZ, loc_30	; until terminating
		clr	cupos		; cursor position
		dec	R13

;------------------------------------------------------------------------------
; USER FUNCTION AND
; via IL_US, RR4=arg1, RR2=arg2
;------------------------------------------------------------------------------

LN3C:		and	R3, R5
		and	R2, R4
		ret

;------------------------------------------------------------------------------
; fetch word (^)
;------------------------------------------------------------------------------

sub_41:		call	loc_47		; hi byte
		ld	R2, R3		; to R2
		db  0Dh			; next code as NOP -> sub_49
					; lo byte to R3

;------------------------------------------------------------------------------
; USER FUNCTION @
; via IL_US, RR4=addr, RR2=value
;------------------------------------------------------------------------------

LN47:
loc_47:		clr	R2		; only 1 byte
sub_49:		inc	R4		; R4=0 (is src a register?)
		djnz	R4, loc_4F	; if 0
		ld	R3, @R5		; load from register
		db  0Dh			; next code as NOP
loc_4F:		ldc	R3, @RR4	; else load from memory
		incw	RR4
		ret

;------------------------------------------------------------------------------
; single character input and output drivers
;------------------------------------------------------------------------------

; input:		tcm	imr, #88h	; Interrupt mask register
; 		jp	LE, 1012h  + start	; external input driver
; input1:		tm	irq, #8		; Interrupt request register
; 		jr	Z, input1	; wait for serial i/o
; 		ld	R3, sio		; in from Serial I/O
; ;
; output:		inc	cupos		; Print	column counter,	cursor position
; 		and	R3, #7Fh
; output1:	tcm	imr, #88h	; Interrupt mask register
; 		jp	LE, 1015h + start	; external output driver
; 		and	irq, #0E7h	; Interrupt request register
; 		ld	sio, R3		; out to Serial I/O
; output2:	tm	irq, #10h	; Interrupt request register
; 		jr	Z, output2	; wait for serial i/o
; 		ret


; Anbindung ES 4.0.
; ggf noch weitere Register sichern?
; %10-1F 	Basic, MON, EDI

input:		call	0C1Dh		; TSK
		ld	R3, 5Ah
		; and echo char:

output:		inc	cupos
		and	R3, #7Fh 
output1:	ld	5Ah, R3
		call	0827h	; ZBS
		ret

;------------------------------------------------------------------------------
; IL instruction: save variable
;------------------------------------------------------------------------------

IL_SV:		call	sub_5FA		; RR4=pop addr from stack
		incw	RR4
		call	loc_81		; save lo value
sub_7F:		ld	R3, R2		; process hi value
;

;------------------------------------------------------------------------------
; LET-PROCEDURE @
; via IL_US, RR4=addr, R3=value
;------------------------------------------------------------------------------

LN81:
loc_81:		inc	R4		; R4=0 (is dest a register) ?
		djnz	R4, loc_87	; if 0
		ld	@R5, R3		; load into register
		db  0Dh			; next code as NOP
loc_87:		ldc	@RR4, R3	; else load into memory RR4=addr, R3=value
		decw	RR4
		ret

;------------------------------------------------------------------------------
; COLD Start
;------------------------------------------------------------------------------

RESET:		
		srp	#0F0h		; MAKE =WORK, SO SHORT
;		clr	imr		; CLEAR INTERRUPT MASK
;		ld	R8, #0B6h	; RAM, FAST, INT STK (10010110)
;		ld	R15, #LBUF	; STACK POINTER
;		ld	R4, #175	; INITIALLY 110 BAUD
;		ld	R5, #0Dh	;  FOR UART (00001101)
;		ld	R7, #41h	; SERIAL I/O (01000001)
;		ld	R1, #3		; START UART CLOCK (%03)
		or	R12, #2		; flags RUN Mode 
		srp	#13h
;		clr	p3		; Port 3
;		ld	R4, #0FFh	; POINT TO SWITCHES
;		ld	R5, #0FDh	; = FFFD
;		ldc	R2, @RR4	; GET THEM
;		ld	R3, #80h	; POSITION SINGLE BIT
;loc_AA:	rl	R3		; (THIS MAY TAKE A WHILE
;		djnz	R2, loc_AA	; BUT WHO CARES?)
;		cp	R3, #20h	; IF NOT 110,
;		jr	Z, loc_B5	; (110 BAUD ALREADY SET)
;		ld	t0, R3		; THEN FIX UP TIMER
;loc_B5:		ei			; *&%$#@ UART
;		tm	p3, #1		; IS BREAK SET?
;		jr	Z, IL_WS
; nondestructively tests memory from low to high addresses.
; Only one byte of every 256 is tested at relative location xxFD (hex).
		ld	R4, #8h-1+hi(start)	; ES 40	test ab 88xxh
loc_BD:		call	test_mem	; test for RAM
		jr	C, loc_D9	; if end of memory reached (FFxx)
		jr	NZ, loc_BD	; while no RAM test next segment
; The first byte of RAM found determines the low boundary of user memory.
; Basic/Debug assumes that if xxFD is RAM, xx00 is also RAM and
; sets pointer register R8-9 to xx00 (hex).
;;		ld	R4, #0e0h
		ld	XX, R4		;
;loc_C6:		call	test_mem	; test for RAM
;		jr	C, loc_CD
;		jr	Z, loc_C6	; while RAM test next segment
;loc_CD:		dec	R4
		ld	R4, #0efh	; ES40	max to EFFF
		ld	YY, R4		; save yy
		ld	bogs_h, R4	; Pointer to bottom of GOSUB stack to yy
		clr	progbeg_l
		ld	bogs_l, #20h	; Base of GOSUB stack = yy20
		db  0Dh			; next code as NOP
loc_D9:		clr	YY		;
		ld	R12, #hi(dcblk)
		ld	R13, #lo(dcblk)
; Automatic start-up allows a program stored in ROM to be
; executed on reset without operator intervention. Automatic
; execution occurs on power-up/reset when the program:
; - is stored in ROM
; - begins at 1020 (hex)
; - begins with a line_h number between 1 and 254 inclusive
		ld	R4, #10h-1  + hi(start)	; -> 1020
		ld	R5, #20h	;
		call	test_mem	; test 1020 for RAM
		jr	Z, IL_MT	; it is RAM
		inc	R2
		djnz	R2, IL_MT	; first Byte = 0 ?
; yes
		ld	progend_h, progbeg_h ; Pointer to end of Basic/Debug program
		ld	progend_l, R5
		ld	progbeg_h, R4	; Pointer to start of Basic program
		ld	progbeg_l, R5

;------------------------------------------------------------------------------
; IL instruction MT: clear program
;------------------------------------------------------------------------------

IL_MT:		ld	R3, #2		; stack reserve
		call	get_cblk	; get from constant block
		ld	R4, progbeg_h	; Pointer to start of Basic program
		ld	R5, progbeg_l
		ld	R2, #0FFh	; BASIC Ende FFFF setzen
		ldc	@RR4, R2
		incw	RR4
		ldc	@RR4, R2
		ldc	R2, @RR4	; second byte <=254 ?
		inc	R2
		jr	NZ, loc_114	; yes -> auto start BASIC program
; no
		add	R3, R5
		adc	R2, R4
		ld	progend_h, R2	; Pointer to end of Basic/Debug	program	in RAM
		ld	progend_l, R3

;------------------------------------------------------------------------------
; IL instruction 2D
; warm start entry point
;------------------------------------------------------------------------------
;
IL_WS:		clr	flags		; Program control flags, disable run mode
loc_114:	srp	#8
		ld	R13, #LBUF	; init nextc to line buffer start
;		ld	p01m, #0B2h	; Ports	0-1 mode, Stack extern
;		inc	R12		; YY=0 ?
;		djnz	R12, loc_125	; Sprung wenn RAM
;		clr	R10		; sonst Arbeit im Registerbreich 00xx
;		ld	R11, #LBUF	; init GOSUB stack
;		ld	p01m, #0B6h	; Ports	0-1 mode, Stack intern
loc_125:	ld	R6, R10		; init togs with bogs (yy20h)
		ld	R7, R11

; enters here to start IL loop;
; return here after error stop
restart_il:	srp	#13h
		ld	sph, togs_h	; Pointer to top of GOSUB stack
		ld	spl, togs_l	; Stack	pointer
		ld	R9, flags	; Program control flags
		and	R9, #2		; run mode?
		ld	R15, #ISTAK0	; init IL Stack
		call	get_cblk0	; get Addr of IL Interpreter
		ld	R8, R3		; R8=hi(il)
		ldc	R3, @RR12	; lo(il)
		add	R9, R3		; RR8 = Adr. IL + 0 (direct mode)/2 (run) = 650/652
		ld	R6, #0FFh
		ld	R7, #0FFh	; Current line number = FFFF

;------------------------------------------------------------------------------
;
;------------------------------------------------------------------------------

il_mainloop:	ldc	R3, @RR8	; fetch	next IL	opcode
		incw	RR8
		ld	R1, #1Fh	; Mask for Branches
		cp	R3, #0Ah
		jr	C, handle_il_SX	; skip to handler for SX instructions
		cp	R3, #30h
		jr	C, handle_0a_2f	; handler for 0A..2F
		cp	R3, #40h
		jr	C, handle_30_3f	; handler for opcodes 30..3F

;------------------------------------------------------------------------------
; handle the opcodes >=0x40
;------------------------------------------------------------------------------

; BR a    40-7F   Relative Branch.
; BC a "xxx"   80xxxxXx-9FxxxxXx  String Not Match Branch.
; BV a    A0-BF   Branch if Not Variable.
; BN a    C0-DF   Branch if Not a Number.
; BE a    E0-FF   Branch if Not Endline.

		clr	R0
		and	R1, R3		; mask out displacement bits
		jr	Z, loc_164	; displacement is zero? yes, skip
;
loc_160:	add	R1, R9		; add displayement to current IL_PC
		adc	R0, R8
loc_164:	and	R3, #0E0h	; calc IL opcode pos: strip bit 4
		swap	R3		; swap nibbles
		add	R3, #4		; add offs to opcode pos
		db  0Dh			; next code as NOP
handle_0a_2f:	add	R3, R3		; in: R3 opcode
		clr	R2
		add	R3, #lo(il_jumptable-4) ; preload address of opcode table - 4
		adc	R2, #hi(il_jumptable-4)
		ldc	R14, @RR2	; fetch addr.
		incw	RR2
		ldc	R3, @RR2
		ld	R2, R14
		call	@RR2		; execute opcode
		srp	#13h
loc_180:	jr	il_mainloop

;------------------------------------------------------------------------------
; IL instruction: advance to next BASIC	line
;------------------------------------------------------------------------------

IL_NX:		tm	flags, #2	; run mode?
		jr	Z, restart_il	; no, warm start
		call	IL_FL		; goto next line
		jr	Z, IL_WS	; if not found
		call	get_cblk0	; get Addr of IL Interpreter
		ld	R8, R3
		ldc	R9, @RR12	; RR8 = Adr. IL
		call	IL_CC		; test for cancel code
		jr	Z, loc_202	; if flag0 = 1, err 0
		ret

;------------------------------------------------------------------------------
; handle J/JS instruction
;------------------------------------------------------------------------------
; JS a    3000-37FF       IL Subroutine Call.
; J  a    3800-3FFF       Jump.
; in: R3 opcode

handle_30_3f:	ldc	R1, @RR8	; get next byte of instruction (low address)
		incw	RR8
		ld	R0, #7
		and	R0, R3		; mask out addressbits
		sub	R0, #4
		add	R3, #8		; JS = 3x, J = 4x
		rrc	R3		; R3 = opcode pos
					; JS => 1, J => 2
		jr	loc_160		; RR0 = displacement

;------------------------------------------------------------------------------
; handle SX instruction (00..09)
; Exchange the top byte of computational stack with that "n" bytes into the stack.
;------------------------------------------------------------------------------
; in: R3 opcode
handle_il_SX:	add	R3, R15		; add opcode to IL Stack Pointer
		ld	R0, @R3		; swap stack values
		ld	R1, @R15
		ld	@R3, R1
		ld	@R15, R0
		jr	loc_180		; return to il_mainloop

;------------------------------------------------------------------------------
; IL instruction return	from IL	call
;------------------------------------------------------------------------------

IL_RT:		pop	R2		; pop return adr
		pop	R3
		pop	R0		; restore IL PC value
		pop	R1
		jr	loc_1CC		; set IL PC

;------------------------------------------------------------------------------
; IL instruction JS a    3000-37FF       IL Subroutine Call.
;------------------------------------------------------------------------------

IL_JS:		pop	R2		; pop return adr
		pop	R3
		push	R9		; save IL PC
		push	R8
		call	sub_1FA		; check for stack overflow
loc_1CC:	call	IL_J		; IL_PC := RR0
loc_1CF:	jr	loc_180

;------------------------------------------------------------------------------
; IL instruction gosub
;------------------------------------------------------------------------------

IL_GS:		call	IL_SP		; pop RR2 from stack (line number)
		pop	R0		; discard return addr (17Eh)
		pop	R1
		push	R3		; push line number on stack
		push	R2
		call	sub_1FA		; check for stack overflow
		jr	loc_225		; reinit togs and goto il_mainloop

;------------------------------------------------------------------------------
; IL instruction string	match branch
; jump forward if string was not matched
;------------------------------------------------------------------------------

IL_BC:		ld	R4, R10		; save pointer to current BASIC character
		ld	R5, R11
bc_loop:	ldc	R2, @RR8	; read char
		incw	RR8
		call	get_nchar	; consume char from BASIC input
		sub	R2, R3		; compare char
		jr	Z, bc_loop	; while equal
		adc	R2, R2		; last char -> bit 7 set, r2=80h -> r2+r2=0
		jr	Z, locret_259	; string matched -> return
		ld	R10, R4		; else restore BASIC pointer
		ld	R11, R5
		jr	IL_FBR		; and branch forward

;------------------------------------------------------------------------------
; check for stack overflow
;------------------------------------------------------------------------------

sub_1FA:	srp	#0Ch
		inc	R12		; test YY=0?
		djnz	R12, loc_204	; no, is RAM System
		cp	spl, #ISTAKL	; else check Stack pointer low boundary
loc_202:	jr	loc_20C
;
loc_204:	cp	sph, R4		; compare Stack pointer with progend
		jr	NZ, loc_20C
		cp	spl, R5
loc_20C:	srp	#13h
		jr	loc_283		; err or return

;------------------------------------------------------------------------------
; IL instruction: restore saved	line
; after RETURN
;------------------------------------------------------------------------------

IL_RS:		pop	R0		; discard return addr
		pop	R1		; (adr. 17Eh)
		cp	sph, bogs_h	; Pointer to bottom of GOSUB stack
		jr	NZ, loc_21C	; GOSUB Stack underflow?
		cp	spl, bogs_l	; 
loc_21C:	jr	NC, error	; yes
		pop	R2		; line number with GOSUB
		pop	R3
		call	expr_push_word	; save again for IL_GO
loc_225:	ld	togs_h, sph	; reinit Pointer to top of GOSUB stack
		ld	togs_l, spl	; 
		jr	loc_1CF		; goto il_mainloop

;------------------------------------------------------------------------------
; IL instruction: duplicate top	of expr_stack
;------------------------------------------------------------------------------

IL_DS:		call	IL_SP		; pop top of expr_stack
		call	expr_push_word	; push twice on expr_stack
		jr	expr_push_word

;------------------------------------------------------------------------------
; IL instruction: push word (hi byte 0)
;------------------------------------------------------------------------------

IL_LN:		call IL_LB

;------------------------------------------------------------------------------
; IL instruction: push byte
;------------------------------------------------------------------------------

IL_LB:		ldc	R3, @RR8	; get byte from IL
		incw	RR8		; Pointer to next IL
		jr	expr_push_byte

;------------------------------------------------------------------------------
; IL instruction: add
;------------------------------------------------------------------------------

IL_AD:		call	sub_5FA		; pop value from Stack
		add	R3, R5		; add to current value
		adc	R2, R4
		jr	expr_push_word	; push result

;------------------------------------------------------------------------------
; long branch instruction
;------------------------------------------------------------------------------

IL_BBR:		sub	R1, #20h	; sub direction offset
		sbc	R0, #0
		db  0Dh			; next code as NOP

; BV, BE, BN: in case of error, go on char back and throw error
loc_24F:	decw	RR10		; Basic Pgm. Pointer

;------------------------------------------------------------------------------
; IL instruction: branch forward
;------------------------------------------------------------------------------

IL_FBR:		or	R0, R0
		jr	Z, error	; was displacement 0?

;------------------------------------------------------------------------------
; IL instruction: IL Jump
;------------------------------------------------------------------------------

IL_J:		ld	R8, R0		; set IL PC
		ld	R9, R1		; to new value
locret_259:	ret

;------------------------------------------------------------------------------
; IL instruction: branch if not	variable
;------------------------------------------------------------------------------

IL_BV:		call	get_nchar	; get current BASIC char
		cp	R3, #'A'
		jr	C, loc_24F	; if no var name,
					; go on char back and throw error
		cp	R3, #'Z'
		jr	UGT, loc_24F	; if no var name
					; go on char back and throw error
		add	R3, R3		; yes, double the ASCII code
					; (make it a word index into var table)
		add	R3, #0A0h	; 20h - 2 * '@'
		ld	R2, YY		; RR2 = addr of var

;------------------------------------------------------------------------------
; push word on expr_stack
;------------------------------------------------------------------------------

expr_push_word:	dec	R15		; push hi byte
		ld	@R15, R2	; after that follows lo byte:

;------------------------------------------------------------------------------
; push byte on expr_stack
;------------------------------------------------------------------------------

expr_push_byte:	dec	R15		; set new stack top
		ld	@R15, R3	; push value
		cp	R15, #ISTAKL	; stack overflow?
		jr	ULE, error
		or	YY, YY		; RAM-less?
		jr	NZ, locret_259	; no, exit
		cp	R15, eol_l	; else: stack underflow?
loc_283:	jr	UGT, locret_259	; no, exit, else error:

;------------------------------------------------------------------------------
; common error routine
;------------------------------------------------------------------------------

error:		call	gl_docancel
		call	get_cblk0	; get Addr of IL Interpreter
		ldc	R2, @RR12
		sub	R9, R2
		sbc	R8, R3		; RR8 = Adr. IL
		ld	R3, R9
		ld	R2, R8
		ld	R15, #ISTAK0	; reset Stack
		call	emit_number
		ld	R9, #lo(at_msg)	; error message
		ld	R8, #hi(at_msg)
		call	IL_PC		; print string "!"
		tm	flags, #2	; is run mode?
		jr	Z, loc_2B0	; no
		call	IL_PC		; yes: print string " AT "
		ld	R3, line_l
		ld	R2, line_h	; Current line number
		call	emit_number	; print it
loc_2B0:	call	IL_NL
		clr	flags		; cancel run mode
		jp	restart_il

;------------------------------------------------------------------------------
; IL instruction: jump if not end of line
;------------------------------------------------------------------------------

IL_BE:		call	get_nchar	; get BASIC char
		jr	Z, loc_30A	; 0 - BASIC line end
		jr	loc_24F		; else go on char back and throw error

;------------------------------------------------------------------------------
; IL instruction: branch if not	number
; if digit, convert this and following digits to number
; and push on expr_stack
;------------------------------------------------------------------------------

IL_BN:		call	get_nchar	; get BASIC char
		jr	GE, loc_24F	; if not a digit,
					; go on char back and throw error
		and	R3, #0Fh
		call	sub_638		; push digit as word on stack
loc_2CA:	call	get_nchar	; get next BASIC char
		jr	GE, loc_30A	; if not a digit, go on char back and exit
		ld	R1, #10
		call	sub_314		; * 10
		jr	loc_2CA		; next digit

;------------------------------------------------------------------------------
; IL instruction: divide
;------------------------------------------------------------------------------

IL_DV:		call	IL_SP		; pop RR2 from stack (divisor)
il_dv0:		clr	R0		; RR0=0 (start value)
		clr	R1
		ld	R14, #11h	; 16 digits
loc_2DF:	sub	R1, R3
		sbc	R0, R2
		jr	NC, loc_2E9	; underflow?
		add	R1, R3		; then readd 
		adc	R0, R2
loc_2E9:	ccf
		rlc	@R15		; on stack: dividend
		inc	R15
		rlc	@R15
		dec	R15
		rlc	R1		; RR0*2
		rlc	R0
		djnz	R14, loc_2DF	; process all digits
		ret

;------------------------------------------------------------------------------
; IL instruction: convert hex number
;------------------------------------------------------------------------------

IL_HN:		call	sub_636		; push inital value 0 on expr_stack
loc_2FB:	call	get_nchar	; next char from basic code
		jr	LT, loc_30D	; if digit
		cp	R3, #'A'
		jr	C, loc_30A	; no hex digit
		sub	R3, #'F'+1	; 'A'..'F' -> FA..FF
		jr	C, loc_30D	; if A..F, else
loc_30A:	decw	RR10		; reset next char pointer
		ret

loc_30D:	ld	R1, #10h
		call	sub_314		; * 16
		jr	loc_2FB		; and next digit

sub_314:	clr	R0
		and	R3, #0Fh	; 0..9->0..9,FA..FF->A..F
		jr	loc_323

;------------------------------------------------------------------------------
; IL instruction multiply
;------------------------------------------------------------------------------

IL_MP:		ld	R1, @R15	; RR0=factor2 from stack
		inc	R15
		ld	R0, @R15
		inc	R15
		clr	R3		; RR2=prod=0 (start value)
loc_323:	clr	R2
		ld	R14, #11h	; fr 16 Bit/Stellen
		rcf			; Carry=0
loc_328:	inc	R15		; factor1 / 2
		rrc	@R15		;
		dec	R15		;
		rrc	@R15		; factor1 lo bit in carry
		jr	NC, loc_335	; wenn lo bit = 0
		add	R3, R1		; sonst aufaddieren
		adc	R2, R0		; RR2 += RR0
loc_335:	rrc	R2		;
		rrc	R3		; RR2 / 2
		djnz	R14, loc_328	; alle Stellen abarbeiten
		ret

;------------------------------------------------------------------------------
; IL instruction PQ
;------------------------------------------------------------------------------

loop_pq:	call	output		; char output
IL_PQ:		scf
		call	get_nchar1	; get next char from BASIC text
		jr	Z, loc_30A	; exit if 00
		cp	R3, #'"'	; is character string terminator?
		jr	Z, loc_374	; yes, exit (and skip next IL)
		jr	loop_pq		; otherwise emit char
					; and redo PQ instruction

;------------------------------------------------------------------------------
; IL instruction goto
;------------------------------------------------------------------------------

IL_GO:		call	loc_5FD		; pop RR4 (lineno) from Stack
		ld	R10, progbeg_h	; Pointer to start of Basic program
		ld	R11, progbeg_l
		ld	R0, R6		; save last line number
		ld	R1, R7
		call	sub_4CF		; get current line number
		jr	loc_35F
;
loc_35C:	call	IL_FL		; goto next line
loc_35F:	cp	R4, R6		; complare with lineno
		jr	NZ, loc_365	; next line
		cp	R5, R7
loc_365:	jr	UGT, loc_35C	; next line while less than lineo
		jr	Z, locret_376	; return if found
		ldc	R14, @RR8	; current IL PC
		cp	R14, #28h	; (28h=id from SN) error before SN-Stament?
		jr	Z, locret_376	; yes
		ld	R7, R1		; else set current line number
		ld	R6, R0		; to found number
; skip next IL
loc_374:	incw	RR8		; Pointer to next IL
locret_376:	ret

;------------------------------------------------------------------------------
; IL instruction: restore BASIC	pointer
;------------------------------------------------------------------------------


IL_RB:		ld	R3, nextc_h	; dest: Pointer to next char
		db  0Dh			; next code as NOP

;------------------------------------------------------------------------------
; IL instruction save BASIC pointer
;------------------------------------------------------------------------------

IL_SB:		ld	R3, R10		; dest: RR10=Basic Pgm. Pointer
		cp	R3, YY		; restore or save?
		jr	NZ, loc_386	; if restore
; save BASIC pointer into nextc
sub_381:	ld	nextc_h, R10
		ld	nextc_l, R11
		ret
; swap BASIC pointer and nextc
loc_386:	ld	R2, nextc_h	; save nextc into RR2
		ld	R3, nextc_l
		call	sub_381		; fill nextc with BASIC pointer
		ld	R10, R2		; set BASIC pointer to old nextc
		ld	R11, R3
		ret

;------------------------------------------------------------------------------
; IL instruction: 0A halt
; skip next instruction if user flag1 is 1
; after IL_NX -> 651h (X1, ENTRY POINT NEXT STMT MODE)
;------------------------------------------------------------------------------

; Whether or not external I/O drivers are enabled, though it
; is of primary use when they are, the program execution can be
; halted by setting user flag 1, which is the least significant bit
; of the flag register (R252), to 1.


IL_CC:		
;		tcm	imr, #88h	; Interrupt mask register
;		jr	LE, loc_3A3	; external driver
;		ld	R3, #4		; cancel code
;		call	get_cblk	; get from constant block (default 1Bh)
;		xor	R3, sio		; compare with Serial I/O
;		add	R3, R3		; ?
;		jr	Z, locret_376	; return, else ?
loc_3A3:	ld	R3, flags	; Program control flags
		or	R3, #0FEh	; Test user flag0
		inc	R3		; User flag0 = 1 -> Z, (error 0)
		jr	NZ, loc_374	; else skip next IL
					; after IL_NX -> 651h
		ret

;------------------------------------------------------------------------------
; IL instruction: Insert BASIC line
;------------------------------------------------------------------------------

; local subroutine, count chars
sub_3AC:	call	sub_49		; R3 = get char from @RR4, incw	RR4
		inc	R14		; inc counter
		inc	R3
		djnz	R3, sub_3AC	; while R3 > 0, get next char
		ret

IL_IL:		ld	R4, nextc_h	; Pointer to next char in input 
		ld	R5, nextc_l	; line buffer, after line number
		clr	R14		; length counter = 0
		call	sub_3AC		; get length of input
		call	IL_SP		; pop RR2 = line number from stack
		ld	R5, R11		; RR4 = Basic Pgm. Pointer
		ld	R4, R10		; RR10 points to follow line (IL_GO)
		decw	RR10		; go to line number pos
		decw	RR10		; (for insert after program)
		scf			; -1
		sbc	R7, R3		; RR6 = current Basic line number
		sbc	R6, R2		; substract insert line number
		ld	R7, R3		; save lo(line number)
		jr	NC, loc_40C	; if current line <> insert line no

		; replace current line
		djnz	R14, loc_3D7	; length = 0 -> delete line
		ld	R14, #-3	; remove 3 Bytes
		ld	R2, #-1		; set delete flag 
loc_3D7:	inc	R14
		com	R14
		call	sub_3AC		; count chars in current line
		com	R14
		jr	Z, loc_435	; if equal space, insert
		jr	PL, loc_412	; if space needed
		ld	R5, R11		; else compress program
		ld	R4, R10
		ld	R0, R4		; RR0 = dest
		ld	R1, R5
		sub	R5, R14
		sbc	R4, #-1		; RR4 = src
loc_3F0:	call	loc_4F		; R3 = @RR4, incw RR4
		ldc	@RR0, R3	; copy to @RR0
		incw	RR0
;;		cp	reg14, progend_h
		cp	r4, progend_h
		jr	C, loc_3F0
;;		cp	reg15, progend_l
		cp	r5, progend_l
		jr	C, loc_3F0	; copy whil not end
		ld	progend_h, R0	; set new progend
		ld	progend_l, R1
		inc	R2		; delete flag set?
		jr	Z, locret_449	; then quit
		dec	R2
		jr	loc_435		; else insert new line

		; insert before current line
loc_40C:	djnz	R14, loc_40F	; if length counter > 0
		ret			; else quit
loc_40F:	inc	R14		; three mory bytes for line no
		inc	R14		;  and end 00
		inc	R14
loc_412:	ld	R5, progend_l
		ld	R4, progend_h
		ld	R1, R5
		ld	R0, R4
		add	R5, R14		; add length of input
		adc	R4, #0
		call	loc_204		; abort if stack overflow
		ld	progend_h, R4	; set new progend
		ld	progend_l, R5
		; make space before current line
		; move RAM from old end to new end
loc_426:	ldc	R3, @RR0
		decw	RR0
		call	loc_87		; @RR4 = R3, decw RR4
		cp	R4, R10
		jr	UGT, loc_426
		cp	R5, R11
		jr	UGT, loc_426
		
		; insert new line
loc_435:	ldc	@RR10, R2	; hi(line number)
		incw	RR10
		ldc	@RR10, R7	; lo(line number)
		ld	R4, nextc_h	; Pointer to next char
		ld	R5, nextc_l	; in input line buffer
loc_43F:	incw	RR10
		call	sub_49		; R3 = @RR4
		ldc	@RR10, R3	; copy chars
		inc	R3
		djnz	R3, loc_43F
locret_449:	ret

;------------------------------------------------------------------------------
; IL instruction: get input line
; uses expr_stack as buffer
;------------------------------------------------------------------------------

IL_GL:		dec	R15
		ld	eol_l, #LBUF-1
		ld	R10, YY		; Pointer to the end of	the line buffer
		ld	R11, #LBUF
loc_453:	inc	eol_l
gl_loop:	call	input		; R3=input char
		ld	R2, R3
		ld	R3, #3		; backspace code
		call	get_cblk	; get from constant block
		cp	R3, R2		; is backspace?
		jr	Z, gl_dobackspace
		ldc	R3, @RR12	; cancel code from constant block
		cp	R3, R2
		jr	Z, gl_docancel
		or	R2, R2		; ignore, if no input char (if 0)
		jr	Z, gl_loop
		cp	R2, #7Fh	; if 0x7F, ignore
		jr	Z, gl_loop
		cp	R2, #0Ah	; if LF, ignore
		jr	Z, gl_loop
loc_477:	ld	R5, eol_l
		ld	R4, eol_h	; Pointer to the end of	the line buffer
		ld	R3, #0Dh	; load CR
		xor	R3, R2		; ENTER? -> 0
		jp	Z, loc_27	; write 0 (end of basic line), exit GL
		call	sub_7F		; else write char in pgm. memory, decw eol
		cp	eol_l, #0E7h	; upper bound input buffer
		jr	UGT, loc_492	; max size of input buffer reached
		inc	R4
		djnz	R4, loc_453	; while eol>0 (no RAM underflow)
		cp	R15, eol_l	; stack collision?
		jr	UGT, loc_453	; no
loc_492:	ld	R3, #7		; emit BEL character (line overflow)
		call	output1
		jr	gl_loop

gl_dobackspace:	dec	eol_l		; one char back
		cp	R11, eol_l	; eol underflow?
		jr	ULE, gl_loop	; no

gl_docancel:	ld	eol_l, #LBUF	; ie no input
		ld	R2, #0Dh	; end of input
		jr	loc_477		; -> exit GL

;------------------------------------------------------------------------------
; IL instruction compare
; stack: RIGHT EXPR, MASK, LEFT EXPR
; compare RIGHT EXPR with LEFT EXPR
; MASK is bit2 = less
;      bit1 = equal
;      bit0 = greater
; if compare result AND mask return <>0, next IL op is skipped
;
;           bit 210
; 1 > 		001
; 2 =		010
; 3 >=		011
; 4 <		100
; 5 <>		101
; 6 <=		110
;------------------------------------------------------------------------------

IL_CP:		call	IL_SP		; pop RIGHT EXPR from stack into RR2
		ld	R14, @R15	; pop MASK from stack
		inc	R15
		sub	R3, @R15	; and subtract RIGHT EXPR - LEFT EXPR
		inc	R15
		sbc	R2, @R15
		jr	LT, cp_is_lt	; RIGHT EXPR < LEFT EXPR
		jr	GT, cp_is_gt	; RIGHT EXPR > LEFT EXPR
		or	R3, R2
		jr	Z, cp_is_eq	; = 0
cp_is_gt:	rr	R14		; bit 2 in carry
cp_is_eq:	rr	R14		; bit 1 in carry
cp_is_lt:	rr	R14		; bit 0 in carry
		jr	NC, loc_4C4	; if mask bit is not set,
		incw	RR8		; skip next IL op
loc_4C4:	ld	R3, @R15
		inc	R15
		ret

;------------------------------------------------------------------------------
; IL instruction : execute program
;------------------------------------------------------------------------------

IL_XQ:		ld	flags, #2	; run mode
		decw	RR10		; set pointer to line number
		decw	RR10

;------------------------------------------------------------------------------
; get BASIC line number
;------------------------------------------------------------------------------

sub_4CF:	ldc	R6, @RR10	; RR6 line number
		incw	RR10
		ldc	R7, @RR10
		incw	RR10
		ld	R14, R6		; test for FFFF (end of basic)
		and	R14, R7
		inc	R14
		ret

;------------------------------------------------------------------------------
; IL instruction 29 next line
;------------------------------------------------------------------------------

IL_FL:		ldc	R3, @RR10	; get current char from basic pgm.
		incw	RR10		; next position
		inc	R3		; test for 00
		djnz	R3, IL_FL	; next while R3 is not 00
		jr	sub_4CF		; get next line number

;------------------------------------------------------------------------------
; IL instruction: print number
;------------------------------------------------------------------------------

IL_PN:		call	IL_SP		; pop TOS from Stack into RR2

;------------------------------------------------------------------------------
; emit number RR2 decimal
;------------------------------------------------------------------------------

emit_number:	ld	R14, #16	; 16 bit
		clr	R1		; inital value
		clr	R0
loc_4EF:	adc	R3, R3		; number * 2
		adc	R2, R2
		adc	R1, R1		; result*2 + hi bit
		da	R1		; to BCD
		adc	R0, R0
		da	R0
		djnz	R14, loc_4EF	; all bits
		dec	R15		; push result on stack
		ld	@R15, R0
		dec	R15
		ld	@R15, R1
		adc	R3, R3		; highest digit zero?
		jr	Z, IL_PH	; then output max. four dec. digits
		call	unk_544		; else out R3 low nibble (highest digit)
		ld	R3, #-18
		inc	R15
		jr	loc_528		; -> xxx_52F, output four digits

;------------------------------------------------------------------------------
; IL instruction: print hex number
; suppress leading zero's
; in: number 16 bit on IL stack
;------------------------------------------------------------------------------

IL_PH:		ld	R3, #-18	; 3*6
		inc	R15		; high byte
loc_514:	ld	R2, #0F0h
		and	R2, @R15	; mask high nibble
		jr	NZ, loc_528	; -> xxx_52F, out 4 digits
					; xxx_541, out 2 digits
		add	R3, #6		;
		jr	Z, loc_528	; Z bei 3x6 -> xxx_547, out 1 digit
		or	R2, @R15	; test 2nd digit
		jr	NZ, loc_528	; -> xxx_535, out 3 digits
		add	R3, #6		; proc offs lower nibble
		djnz	R15, loc_514	; lo byte
;
loc_528:	add	R3, #lo(xxx_541)
		ld	R2, #hi(xxx_541)
		jp	@RR2		; 52F, 535, 541

; out 4 digits
xxx_52F:	nop
		ld	R3, @R15
		call	sub_545		; out high nibble
; out 3 digits
xxx_535:	ld	R3, @R15
		call	unk_544		; out low nibble
		dec	R15		; lo byte
; out 2 digits
		ld	R3, @R15
		call	sub_545		; out high nibble, then lo nibble
; out 1 digit
xxx_541:	call	IL_SP		; pop number from stack
; out low nibble
unk_544:	db  0Dh			; next code as NOP
; out high nibble
sub_545:	swap	R3		; swap nibbles
xxx_547:	and	R3, #0Fh
		ld	R2, #9
		cp	R2, R3
		adc	R3, #'0'
		da	R3
		jp	output

;------------------------------------------------------------------------------
; IL instruction print string
; last byte has the most significant bit set to one
;------------------------------------------------------------------------------

IL_PC:		ldc	R3, @RR8
		call	output
		ldc	R3, @RR8
		incw	RR8
		inc	R3		; 0...7F ?
		jr	PL, IL_PC	; while ASCII char
		ret

;------------------------------------------------------------------------------
; Interpreter jump table
;------------------------------------------------------------------------------
; must be in a segment!

; Interpreter jump table									; 6502 vs. z8671 IL
il_jumptable:	dw IL_JS		; 30-37 IL Subroutine Call
		dw IL_J			; 38-3F IL Jump
		dw IL_BBR		; 40-5F backward branch
		dw IL_FBR		; 60-7F	forward_branch
		dw IL_BC		; 80-9F	"xxX" string NOT match branch
		dw IL_BV		; A0-BF	branch if NOT variable
		dw IL_BN		; C0-DF branch if NOT number
		dw IL_BE		; E0-FF	branch if NOT eoln
		dw IL_CC		; 0A	stop if cancel code
		dw IL_LB		; 0B x	push literal byte to TOS			; 6502 code 09
		dw IL_LN		; 0C x	push addr (00xxh) to TOS			; 6502 code 0a only lo byte
		dw IL_DS		; 0D	duplicate stack top				; 6502 code 0b
		dw IL_SP  		; 0E	pop TOS						; 6502 code 0c
		dw IL_TS		; 0F	test sign; skip next IL if TOS is positive 	; 6502 reserved
		dw IL_SB		; 10	save BASIC pointer
		dw IL_RB		; 11	restore BASIC pointer
		dw IL_FV		; 12	fetch variable
		dw IL_SV 		; 13	store variable
		dw IL_GS		; 14	save GOSUB line
		dw IL_RS		; 15	restore saved line
		dw IL_GO		; 16	goto
		dw IL_NE		; 17	negate
		dw IL_AD  		; 18	add
		dw 1018h   + start		; 19	reserved					; 6502 substract
		dw IL_MP		; 1A	multiply
		dw IL_DV		; 1B	divide
		dw IL_CP		; 1C	compare
		dw IL_NX		; 1D	next BASIC statement
		dw 101Bh  + start		; 1E	reserved
		dw 101Eh  + start 		; 1F	reserved					; 6502 list program
		dw IL_PN		; 20	print number
		dw IL_PQ		; 21	print BASIC string
		dw IL_PT		; 22	print tab
		dw IL_NL 		; 23	new line
		dw IL_PC		; 24 "xx" print literal string
		dw IL_PH		; 25	print hex number				; 6502 reserved
		dw IL_HN		; 26	convert hex number				; 6502 reserved
		dw IL_GL		; 27	get input line
		dw IL_SN		; 28	put current line number on stack		; 6502 reserved
		dw IL_FL		; 29	follow line; goto next line			; 6502 reserved
		dw IL_IL		; 2A	Insert BASIC line
		dw IL_MT 		; 2B	mark BASIC program space empty
		dw IL_XQ		; 2C	execute
		dw IL_WS		; 2D	stop execution
		dw IL_US		; 2E	machine language subroutine call
		dw IL_RT		; 2F	IL subroutine return

;------------------------------------------------------------------------------
; test if its RAM
;------------------------------------------------------------------------------

; Test (R4+1)R5 for RAM
test_mem:	scf			; C = ram end
		inc	R4		; R4+1
		jr	Z, locret_5D2	; exit on FFxx RAM end

;es40
;		cp	r4,#0F0h	
;		jr	Z, locret_5D2	; exit on EFxx RAM end
			
		ldc	R3, @RR4	; get byte
		com	R3
		ldc	@RR4, R3	; store complement
		com	R3
		ldc	R2, @RR4	; get byte
		ldc	@RR4, R3
		adc	R3, R2		; on RAM result is 0
		rcf
locret_5D2:	ret			; Z = RAM, NZ = ROM

;------------------------------------------------------------------------------
; get next BASIC char (without advance)
;------------------------------------------------------------------------------

get_nchar:	rcf
get_nchar1:	inc	R10		; r10=0 ?
		djnz	R10, get_nchar2	; no
		ld	R3, @R11	; yes
		db  0Dh			; next code as NOP
get_nchar2:	ldc	R3, @RR10
		incw	RR10
		jr	C, get_nchar4
		cp	R3, #' '	; is it a space?
		jr	Z, get_nchar	; yes, skip that
		cp	R3, #60h	; is it lower case?
		jr	C, get_nchar3
		sub	R3, #20h	; yes, to upper
get_nchar3:	cp	R3, #'0'	; is it number?
		jr	LT, get_nchar4
		cp	R3, #'9'+1
		jr	LT, get_nchar5
get_nchar4:	or	R3, R3		; 00 ?
get_nchar5:	ret

;------------------------------------------------------------------------------
; pop two arguments from stack
;------------------------------------------------------------------------------

sub_5FA:	call	IL_SP		; RR2 = pop of expr_stack
loc_5FD:	ld	R5, @R15	; RR4 = pop next argument from stack
		inc	R15
		ld	R4, @R15
		inc	R15
		ret

;------------------------------------------------------------------------------
; IL instruction US: machine language subroutine call
;------------------------------------------------------------------------------

IL_US:		call	sub_5FA		; pop value and addr from stack
		ld	R1, @R15	; pop function addr from stack
		inc	R15
		ld	R0, @R15
		inc	R15
		call	@RR0		; call function
		srp	#13h
		jr	loc_633

;------------------------------------------------------------------------------
; IL instruction:  skip next IL op if TOS is positive
;------------------------------------------------------------------------------

IL_TS:		call	IL_NE		; negate
		or	R2, R2		; test sign
		jr	PL, IL_NE	; if plus, negate back
		incw	RR8		; else skip next IL op

;------------------------------------------------------------------------------
; IL instruction: negate top of	stack
;------------------------------------------------------------------------------

IL_NE:		call	IL_SP		; pop RR2 from stack
		com	R2		; negate
		com	R3
		incw	RR2
		jr	loc_633		; push RR2 on expr_stack

;------------------------------------------------------------------------------
; IL instruction: fetch	variable
;------------------------------------------------------------------------------

IL_FV:		call	loc_5FD		; pop RR4 from Stack (var addr)
		call	sub_41		; fetch word
		jr	loc_633		; push RR2 on expr_stack (value)

;------------------------------------------------------------------------------
; IL instruction: push current line number on stack
;------------------------------------------------------------------------------

IL_SN:		ld	R2, R6		; RR6 = line number
		ld	R3, R7
loc_633:	jp	expr_push_word	; push RR2 on expr_stack

;------------------------------------------------------------------------------
; push 0 on expr_stack
;------------------------------------------------------------------------------

sub_636:	clr	R3
sub_638:	clr	R2
		jr	loc_633		; push 0 on expr_stack

;------------------------------------------------------------------------------
; get constant block
; in: R3 - idx in constant block
;------------------------------------------------------------------------------

; get constant block
get_cblk0:	clr	R3		; R3 = 0
; get idx R3 from constant block
get_cblk:	and	R13, #0F0h
		or	R13, R3
; get next byte from constant block
get_cblk1:	ldc	R3, @RR12
		inc	R13
		cp	R3, #0FEh	; sequence end? (this is code 'inc R15', loc_27)
		ret


;------------------------------------------------------------------------------
; error messages
;------------------------------------------------------------------------------

at_msg:		db  '!', 07h + 80h	; BELL
at_msg2:	db  " AT", ' ' + 80h

;------------------------------------------------------------------------------
; The IL interpreter
; see tb_il_z8671
;------------------------------------------------------------------------------
; il     - start direct mode (X1) used in IL_WS (offs is calculated)
; il+1   - next stmt         (X2) used in IL_NX (indirect via il_0a, offs is calculated)
; il+2   - start run mode    (X3) used in IL_WS (offs is calculated)

il:
	section IL
	MACEXP off		; switch on to see the IL binary code
	include	z8671_il_macros.asm
	include	z8671_tbil_es1988.asm
	endsection

; for tb error numbers see output from pass 1

;------------------------------------------------------------------------------
; Registers, RAM etc.
;------------------------------------------------------------------------------

; Internal processor memory & SFR
p0 equ	0				; Port 0
p1 equ	1				; Port 1
p2 equ	2				; Port 2
p3 equ	3				; Port 3

sio equ 0F0h				; Serial I/O
tmr equ 0F1h				; Timer	mode
t1 equ	0F2h				; Timer/counter	1
pre1 equ 0F3h				; t1 prescaler
t0 equ	0F4h				; Timer/counter	0
pre0 equ 0F5h				; t0 prescaler
p2m equ 0F6h				; Port 2 mode
p3m equ 0F7h				; Port 3 mode
p01m equ 0F8h				; Ports	0-1 mode
ipr equ 0F9h				; Interrupt priority register
irq equ 0FAh				; Interrupt request register
imr equ 0FBh				; Interrupt mask register
flags equ 0FCh				; Program control flags
rp equ	0FDh				; Register pointer
sph equ 0FEh				; Stack pointer high
spl equ 0FFh				; Stack pointer low


; The Top Page of RAM-System
;---------------------------

; yy = hi(highest RAM), if no RAM, yy=0 means internal registers
; xx = hi(BASIC-Pgm)

; yyF1 - yyFF Unused.
; yy68 - yyF0 Input line buffer, used for editing in immediate mode and
;		user response to IN or INPUT request in run mode.
; yy56 - yy67 Unused.
; yy54 - yy55 Storage for variable z.
; yy53 - yy52 Storage for variable Y.
; ..
; yy21 - yy22 Storage for variable A.
; yy20	Base of GOSUB stack. Stack grows down to lower memory addresses,
;	and may extend until it reaches the top of the user's Basic/Debug program.

; Pointer Registers - RAM System

; 04-0S => moves up from xx00    High boundary of user program plus stack reserve.
; 06-07 => moves down from yy20  Low boundary and top of GOSUB stack.
; 08-09 => xx00                  Bottom of user memory; first line of user program.
; 0A-0B => yy20                  Top of user memory, high boundary of GOSUB stack.
;				 Initially set to yy20 of high page of RAM.
; 0C-0D => yy68 to yyF0          Last character entered in line buffer. Backspace
;				 subtracts one from this pointer; escape resets it
; 				 to the beginning of the buffer. Rl2 is the page
;				 number for variables and the input buffer.
; 0E-0F => yy68 to yyF0	         Next value to be used from line buffer. INPUT
;				 command resets to the beginning of the buffer;
;				 IN uses all values in the buffer before resetting.

; Register Map for No-RAM System
;-------------------------------

progend_h	equ 4			; Pointer to end of Basic/Debug	program	in RAM
progend_l	equ 5			; 04-05 No-RAM System: Free for USR subroutine
togs_h		equ 6			; Pointer to top of GOSUB stack.
togs_l		equ 7
progbeg_h	equ 8			; Pointer to start of Basic program.
XX		equ progbeg_h
progbeg_l 	equ 9			; No-RAM: progbeg 1020h for auto-start
bogs_h		equ 0Ah			; Pointer to bottom of GOSUB stack
bogs_l		equ 0Bh
eol_h		equ 0Ch			; Pointer to the end of	the line buffer
YY 		equ eol_h
eol_l		equ 0Dh
nextc_h		equ 0Eh			; Pointer to next character to be used in input buffer.
nextc_l		equ 0Fh
; 10-11 Basic/Debug uses as scratch. USR subroutine may use, but can not save values here.
; 12-13 Last argument and result in USR subroutine call.
; 14-15 Second argument in three argument USR subroutine call
reg14		equ 14h			; used in IL_IL as absolute registers
reg15		equ 15h
line_h		equ 16h			; Current line number
line_l		equ 17h			; used in error as absolute registers
; 18-lB Internal variables. Do not modify.
; 1C-1D	Pointer to constant block.
; 1E Basic/Debug uses as scratch. USR subroutine may use, but cannot save values here.
; 1F Internal variable. Do not modify.
cupos		equ 20h			; Print	column counter,	cursor position
; 21 Free register, available for USR subroutine.
; 22-55 Variables A through Z.
; 40-55 Area shared by variables M-Z and GOSUB stack. Variables are destroyed
; 	if stack grows into this range.
; 40-67 GOSUB stack; grows down.
LBUF		equ	68h		; Line Buffer Start
;;ISTAK0		equ	80h		; internal stack End
; 68-7F The Expression Evaluation stack grows from 7F (hex) down,
; 	and the line buffer grows from 68 (hex) up.

ISTAKL		equ	30h		; internal stack End
ISTAK0		equ	50h		; internal stack Start

;------------------------------------------------------------------------------
; SAVE/LOAD f. ES 1988
;------------------------------------------------------------------------------

;%50-51	Programmstart fr SAVE/LOAD	LOW BYTE FIRST!
;%52-53	Programmende fr SAVE		LOW BYTE FIRST!
;%268F Magnetbandausgabe RP = #%60
;%27D1 Magnetbandeingabe RP = #%60

; wie in 4k1_teil1:
CLS		equ	08DDh
ReadBlock	equ	2782h
ES_Save		equ	26ADh

load		srp	#40h		; rcksetzen nicht noetig, macht USR()
		;;push	r0		; CUPOS
		di
		call load1		; lesen
		ei
load0		call	searchend	; rr2 = programmende
		add	r3,#20h		; 0020 fr Stackreserve addieren
		adc	r2,#0
		ld	progend_h,r2	; progend setzen
		ld	progend_l,r3
		;;pop	r0
		ret
;


		ld	r0, progbeg_h	; Pointer to start of Basic program
		ld	r1, progbeg_l
		ld R6, #0		; erster erwarteter Block -1
		di

load1		ld R0, #0FCh		; Pufferadresse
		ld R1, #0
		ld R6, #0		; erster erwarteter Block 
		call ReadBlock		; lesen
		ld	r0, progbeg_h	; Pointer to start of Basic program
		ld	r1, progbeg_l
load2		call ReadBlock		; lesen
		jr Load2		; bei Dateieende wird 1x Ret-Adr. vom Stack genommen
					; dadurch Rckkehr zum Pgm
;
save		srp	#50h
		;;push	r0		; CUPOS
		call	searchend	; rr2 = programmende
		;
		ld	r4,r2		; tausch r2,r3
		ld	r2,r3
		ld	r3,r4
		srp	#60h
		call 	CLS
		call	ES_Save		; SAVE
		ei
		;;pop	r0
		ret

searchend	ld	r0, progbeg_l	; Pointer to start of Basic program
		ld	r1, progbeg_h
		ld	r2, r1
		ld	r3, r0
sav1		ldc	r4,@rr2
		incw	rr2
		;;cp	r4,#0FFh	; test auf FF
		inc	r4
		jr	nz, sav1
		incw	rr2		; rr2 = programmende
		ret

;------------------------------------------------------------------------------
; 
;------------------------------------------------------------------------------

		if 1=0
; Basic-Pgm
	org 9020h
;	db	0,5,"@251=0:REM EXTERNAL IO",0
	db	0,10,"LET A=%1234:PRINT HEX(A)",0
	db	0,20,"LET A=%123:PRINT HEX(A)",0
	db	0,30,"LET A=%12:PRINT HEX(A)",0
	db	0,40,"LET A=%1:PRINT HEX(A)",0
	db	1,0,"GOTO 10",0
	db	0ffh,0ffh

		endif
		
		end start
